from menpofit.result import ParametricAlgorithmResult, MultiFitterResult


# TODO: document me!
class CLMAlgorithmResult(ParametricAlgorithmResult):
    r"""
    """


# TODO: document me!
class CLMFitterResult(MultiFitterResult):
    r"""
    """

